unit Main;

//Program: InvtPend.exe
//Version: 10/11/2009
//Authors: William T. Powers and Bruce B. Abbott
{
Copyright (c) 2009, William T. Powers and Bruce B. Abbott.

====================================================================
SOURCE LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses
====================================================================
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Spin, Menus;

type
  TMainForm = class(TForm)
    ControlPanel: TGroupBox;
    Label6: TLabel;
    ProgramBox: TGroupBox;
    AboutBtn: TButton;
    ExitBtn: TButton;
    LegendBox: TPanel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    DisplayPanel: TGroupBox;
    Panel2: TPanel;
    Panel3: TPanel;
    Display: TPaintBox;
    Panel4: TPanel;
    GraphDisplay: TPaintBox;
    Panel5: TPanel;
    ResetBtn: TButton;
    OptionsBox: TGroupBox;
    ControlTogBtn: TButton;
    ForceTogBtn: TButton;
    Panel1: TPanel;
    Label5: TLabel;
    SpinEditImpL: TSpinEdit;
    ImpulseBtn: TButton;
    Panel6: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure DisplayMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ControlTogBtnClick(Sender: TObject);
    procedure ForceTogBtnClick(Sender: TObject);
    procedure ImpulseBtnClick(Sender: TObject);
    procedure SpinEditImpLChange(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure ExitBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure IdleLoop(Sender: TObject; var Done: boolean);
    procedure ShowPendulum;
    procedure ShowRef;
  end;

var
  MainForm: TMainForm;
  MouseButtonDown: Boolean;

implementation

uses About;

{$R *.dfm}
{$J+}

type
  ObjectType = record
                 X, Y, Fx, Fy, Vx, Vy, Mass: Double;
               end;
  CsysType = record
               P, R, G, E, O: Double;
             end;

const
  XScaleFactor = 200.0;
  YScaleFactor = 200.0;

 var
   Bob,Cart: ObjectType;
   DisplayBitmap, GraphBitmap: TBitmap;
   CartPos, CartVel, BobPos, BobVel, BobAccel: CsysType;
   DeltaX, DeltaY, dt, Impulse, L, L0, F, G, Ke: Double;
   Iters: LongInt;
   Time, ImpulseLength, Stage: Integer;
   HCenter, VCenter, gHCenter, gVCenter, MouseX, MouseY: Integer;
   ImpulseOn, Control: Boolean;
   T1, T2, Freq: Int64;

{----- World-to/from-Screen Mapping Functions ----}

function XtoScreenX(X: double): Integer;
begin
  Result := HCenter + Round(X * XScaleFactor);
end;

function YtoScreenY(Y: double): Integer;
begin
  Result := VCenter - Round(Y * YScaleFactor);
end;

function ScreenXtoX(ScreenX: Integer): double;
begin
  Result := ScreenX/XScaleFactor;
end;

{----- Simulation Initialization -----}

procedure InitPendulum;
begin
  dt := 1e-4;   { TimeVar increment per iteration, sec }
  Ke := 1e7;    { Rod compression force, newtons/meter of compression}
  L0 := 1.0;    { Rod-length, meters }
  G  := 9.8;    { Gravity, meter/sec^2 acceleration}
  Iters := 0;
  Control := True;
//  ImpulseOn := False;
  with Bob do
  begin
    Vx := 0.0;  { m/sec }
    Vy := 0.0;
    X := L0*Sin(0);
    Y := L0*Cos(0);  {meter}
    Mass := 1.0; { kg }
  end;
  with Cart do
  begin
    Vx := 0.0;
    Vy := 0.0;
    X := 0.0;
    Y := 0.0;
    Mass := 0.5; { kg }
  end;
end;

procedure InitControlSystems;
begin
  with CartVel do
    begin
      G := 100.0;
      R :=   0.0;
      E :=   0.0;
      O :=   0.0;
    end;
  with CartPos do
    begin
      G := 50.0;
      R :=  0.0;
      E :=  0.0;
      O :=  0.0;
    end;
  with BobAccel do
    begin
      G := 50.0;
      R :=  0.0;
      E :=  0.0;
      O :=  0.0;
    end;
  with BobVel do
    begin
      G :=  0.5;
      R :=  0.0;
      E :=  0.0;
      O :=  0.0;
    end;
  with BobPos do
    begin
      G := 1.0;
      R := 0.0;
      E := 0.0;
      O := 0.0;
    end;
end;

{----- Control Systems and Dynamics -----}

procedure Dynamics;
begin
  with Bob do
  begin
    DeltaX := Bob.X - Cart.X;
    DeltaY := Bob.Y - Cart.Y;
    L := Sqrt(Sqr(DeltaX) + Sqr(DeltaY)); //Computed length of pole
    F := (L0 - l)*Ke;
    Fx := F*DeltaX/l + Impulse; //Horizontal force on bob
    Fy := F*DeltaY/l - Mass*G;  //Vertical force on bob
    Vx := Vx + Fx/Mass*dt; //Horizontal velocity
    X := X + (Vx + 0.5*Fx*dt/Mass)*dt; //Horizontal position
    Vy := Vy + Fy/Mass*dt; //Vertical velocity
    Y := Y + (Vy + 0.5*Fy*dt/Mass)*dt; //Vertical position
  end;
  with Cart do
  begin
    if Control then Fx := -Bob.Fx + CartVel.O //Horizontal force on cart
    else Fx := -Bob.Fx;
    Vx := Vx + Fx/Mass*dt;
    X := X + (Vx + 0.5*Fx/Mass*dt)*dt; //Horizontal position
  end;
end;

procedure CartVelCont;
begin
  with CartVel do
  begin
    P := Cart.Vx;
    R := CartPos.O; //Proportional to cart position error
    E := R - P;
    O := O + (100.0*E - O)/2.0; //Horizontal force on cart
  end;
end;

procedure CartPosCont;
begin
  with CartPos do
  begin
    P := Cart.X;
    R := BobAccel.o; //Proportional to error in cart pos relative to bob pos
    E := R - P;
    O := 50*E; //Connects to reference for cart velocity
  end
end;

procedure BobAccelCont;
begin
  with BobAccel do
  begin
    P := Bob.x - Cart.x;
    R := BobVel.O; //Proportional to error in bob horiz. velocity
    E := R - P;
    O := -50*E;  //Connects to reference for cart position
  end
end;

procedure BobVelCont;
begin
  with BobVel do
  begin
    P := Bob.Vx;
    R := BobPos.O; //Proportional to error in bob position
    E := R - P;
    O := O + 10.0*(0.5*E - O)*dt; //Connects to reference for bob accel.
    if O > 0.4 then O := 0.4
    else if O < -0.4 then O := -0.4;
  end;
end;

procedure BobPosCont;
begin
  with BobPos do
  begin
    R := ScreenXtoX(MouseX); //Mouse position
    P := Bob.X;
    E := R - P;
    O := 2*E;  //Connects to reference for bob velocity
  end;
end;

{----- Graphics Display -----}

procedure TMainForm.ShowPendulum;
const
  OldX: Integer = 0;
  OldY: Integer = 0;
  OldCx: Integer = 0;
  OldCy: Integer = 0;

  procedure DrawCart(DrawColor: Integer);
  begin
    with DisplayBitmap.Canvas do
    begin
      Pen.Color := DrawColor;
      { Rod }
      MoveTo(OldCx, OldCy);
      LineTo(OldX, OldY);
      { Bob }
      Ellipse(OldX-7, OldY-7, OldX+7, OldY+7);
      { Truck body}
      Rectangle(OldCx-30, VCenter, OldCx + 30, VCenter-15);
      { Left wheel }
      Ellipse(OldCx-30, OldCy-7, OldCx-16, OldCy+7);
      { Right wheel }
      Ellipse(OldCx+30, OldCy-7, OldCx+16, OldCy+7);
    end;
  end;

begin
  { Erase Previous Figure }
  DrawCart(clWhite);
  { Get current bob and cart positions }
  OldX  := XtoScreenX(Bob.X);
  OldY  := YtoScreenY(Bob.Y);
  OldCx := XtoScreenX(Cart.X);
  OldCy := YtoScreenY(Cart.Y);
  { Draw bob and cart in current positions }
  DrawCart(clBlack);
  with DisplayBitmap do { Draw ground }
  begin
    Canvas.MoveTo(0, VCenter+7);
    Canvas.LineTo(Width, VCenter+7);
  end;
end;

procedure TMainForm.ShowRef;
const
  OldRef: Integer = 0;
begin
  with DisplayBitmap.Canvas do
  begin
    { Erase at former position }
    Pen.Color := clWhite;
    MoveTo(OldRef, 25);
    LineTo(OldRef, 35);
    MoveTo(OldRef-1, 25);
    LineTo(OldRef-1, 35);
    { Redraw at new position }
    OldRef := XtoScreenX(BobPos.R);{HCenter + MouseX;}
    Pen.Color := clRed;
    MoveTo(OldRef, 25);
    LineTo(OldRef, 35);
    MoveTo(OldRef-1, 25);
    LineTo(OldRef-1, 35);
  end;
end;

{----- Main Initialization -----}

procedure TMainForm.FormCreate(Sender: TObject);
begin
  DisplayBitmap := TBitmap.Create;
  GraphBitmap := TBitmap.Create;
  MouseX := 0;
  MouseY := 0;
  Iters := 0;
  Stage := 0;
  InitPendulum;
  L := L0;      {L-nought, not 10.0}
  ImpulseLength := 1000;
  ShowPendulum;
  ShowRef;
  QueryPerformanceFrequency(Freq);
  QueryPerformanceCounter(T1);
  Application.OnIdle := IdleLoop;
end;

{----- Simulation Loop -----}

procedure TMainForm.IdleLoop(Sender: TObject; var Done: boolean);
const
  OldRef: Integer = 60;
  OldBob: Integer = 60;
       X: Integer = 0;
       Y: Integer = 0;
 OldCart: Integer = 60;
    OldX: Integer = 0;

begin
  Done := False;
  Case Stage of
    0: begin
         with DisplayBitmap, Canvas do
         begin
           Height := Display.Height;
           Width := Display.Width;
           Brush.Color := clWhite;
           Pen.Width := 3;
           HCenter := Width div 2;
           VCenter := (Height div 2) + 30;
           FillRect(Rect(0, 0, Width, Height));
         end;
         with GraphBitmap, Canvas do
         begin
           Height := GraphDisplay.Height;
           Width := GraphDisplay.Width;
           Brush.Color := clWhite;
           Pen.Width := 3;
           gHCenter := Width div 2;
           gVCenter := Height div 2;
           FillRect(Rect(0, 0, Width, Height));
         end;
         Stage := 1;
       end;
    1: begin
         Dynamics;
         if Control then
         begin
           BobPosCont;
           BobVelCont;
           BobAccelCont;
           CartPosCont;
           CartVelCont;
         end;
         if (Iters mod 300) = 0 then
         begin
           ShowPendulum;
           ShowRef;
           with GraphBitmap, Canvas do
           begin
             { Bob Position Reference }
             Pen.Color := clRed;
             MoveTo(OldX, OldRef);
             OldRef := gVCenter  - Round(25.0*BobPos.R);
             LineTo(OldX+2, OldRef);
             { Cart Position }
             Pen.Color := clBlue;
             MoveTo(OldX,OldCart);
             OldCart := gVCenter - Round(25.0*Cart.X);
             LineTo(OldX+2, OldCart);
             { Bob Position }
             MoveTo(OldX, OldBob);
             OldBob := gVCenter  - Round(25.0*Bob.X);
             Pen.Color := clGreen;
             LineTo(OldX+2, OldBob);
             Inc(OldX);
             if OldX = ClientWidth then
             begin
               FillRect(Rect(0, 0, ClientWidth, ClientHeight));
               Iters := 0;
               OldX := 0;
             end;
           end;
           Display.Canvas.Draw(0, 0, DisplayBitmap);
           GraphDisplay.Canvas.Draw(0, 0, GraphBitmap);
         end; {if Iters}
         with DisplayBitmap.Canvas do
         begin
           if Time = ImpulseLength then
           begin
             Pen.Color := clRed;
             X := XtoScreenX(Bob.X);
             Y := YtoScreenY(Bob.Y);
             MoveTo(X-50, Y);
             LineTo(X-10, Y);
           end;
           if Time = 1 then
           begin
             Pen.Color := clWhite;
             MoveTo(X-50, Y);
             LineTo(X-10, Y);
             Impulse := 0.0;
           end;
         end;
         if Time > 0 then Dec(Time);
         Inc(Iters);
       end; {Stage 1}
  end; {Case Stage of}
  repeat
    QueryPerformanceCounter(T2);
    Application.ProcessMessages;
  until T2 - T1 >= Freq/6000;
  T1 := T2;
end;

{----- Event Handlers -----}

procedure TMainForm.DisplayMouseMove(Sender: TObject; Shift: TShiftState;
                                     X, Y: Integer);
var
  LeftButtonDown: Boolean;
begin
  LeftButtonDown := ssLeft in Shift;
  if  {MouseButtonDown} LeftButtonDown then
    begin
      MouseX := X - HCenter;
      MouseY := Y;
    end;
end;

procedure TMainForm.ControlTogBtnClick(Sender: TObject);
begin
  if Control then
    begin
      Control := False;
      ControlTogBtn.Caption := 'Turn &Control ON';
    end
  else
    begin
      Control := True;
      ControlTogBtn.Caption := 'Turn &Control OFF';
    end;
end;

procedure TMainForm.ForceTogBtnClick(Sender: TObject);
begin
  if Impulse = 0.0 then
    begin
      ForceTogBtn.Caption := 'Turn &Force OFF';
      Impulse := 1.0
    end
  else
    begin
      ForceTogBtn.Caption := 'Turn &Force ON';
      Impulse := 0.0;
    end
end;

procedure TMainForm.ImpulseBtnClick(Sender: TObject);
begin
  if Impulse = 0.0 then // prevent if Force is On
  begin
    Impulse := 2.0;
    Time := ImpulseLength;
  end;
end;

procedure TMainForm.SpinEditImpLChange(Sender: TObject);
begin
  ImpulseLength := SpinEditImpL.Value;
end;

procedure TMainForm.ResetBtnClick(Sender: TObject);
begin
  with DisplayBitmap, Canvas do
    FillRect(Rect(0, 0, Width, Height));
  InitPendulum;
  SpinEditImpL.Value := 500;
  MouseX := 0;
  Control := True;
  ControlTogBtn.Caption := 'Turn &Control OFF';
  ForceTogBtn.Caption := 'Turn &Force ON';
  InitControlSystems;
end;

procedure TMainForm.FormPaint(Sender: TObject);
begin
  Display.Canvas.Draw(0, 0, DisplayBitmap);
  GraphDisplay.Canvas.Draw(0, 0, GraphBitmap);
end;

procedure TMainForm.ExitBtnClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  GraphBitmap.Free;
  DisplayBitmap.Free;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  with DisplayBitmap, Canvas do
  begin
    Height := Display.Height;
    Width := Display.Width;
    HCenter := Width div 2;
    VCenter := (Height div 2) + 100;
    FillRect(Rect(0, 0, Width, Height));
  end;
  with GraphBitmap, Canvas do
  begin
    Height := GraphDisplay.Height;
    Width := GraphDisplay.Width;
    gHCenter := Width div 2;
    gVCenter := Height div 2;
    FillRect(Rect(0, 0, Width, Height));
  end;
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  AboutBox.ShowModal;
end;

end.
